from __future__ import absolute_import, division, print_function, unicode_literals
from operator import itemgetter
import os, sys, json, logging, time, threading, warnings, collections

try:
    import queue
except ImportError:
    import Queue as queue

import boto3
import boto3.session
from botocore.exceptions import ClientError, ProfileNotFound, NoRegionError

handler_base_class = logging.Handler
logger_base_class = logging.getLogger(__name__)
logging_levels = {'debug': logging.DEBUG, 'info': logging.INFO, 'warning': logging.WARNING, 'error': logging.ERROR,
                  'critical': logging.CRITICAL}


class HyperionLogger():
    """
    Logger object with built in function for setting the handler, level and formatter
    """

    def __init__(self, default_level='info'):
        self.log = None
        self.stream_name = None
        self.log_group = None
        self.default_level = default_level
        self.handler = None
        self.log_format = '%(asctime)s: %(funcName)s: %(levelname)s %(message)s'
        self.formatter = logging.Formatter(self.log_format)
        self.__get_logger__()

    def __get_logger__(self):
        """
        function to grab a logger object.
        :return: logger object
        """

        try:
            self.log = logging.getLogger(__name__)
            self.log.setLevel(logging_levels[self.default_level.lower()])
            warnings.warn(
                'Handler needs to be setup with a stream_name and log group. \n IE: log.add_handler(stream_name,log_group)',
                HyperionWarning)
        except Exception as e:
            warnings.warn('unable to create logger because of {}'.format(e), HyperionWarning)

    def add_handler(self, stream_name, log_group):
        """
        Function to add cloudwatch handler to logger object.
        :param stream_name:
        :param log_group:
        :return: None
        """
        try:
            self.handler = CloudWatchLogHandler()
            self.handler.create_session(profile_name='093937234853-GR_GG_COF_AWS_CoreAdmin')
            self.log_group = log_group
            self.stream_name = stream_name
            self.handler.log_group = self.log_group
            self.handler.create_log_group()
            self.handler.stream_name = self.stream_name
            self.handler.setFormatter(self.formatter)
            self.log.addHandler(self.handler)
            warnings.warn('Logger sending logs to {0} inside log group {1}.'.format(self.stream_name, self.log_group),
                          HyperionWarning)
        except Exception as e:
            warnings.warn('Adding the handler failed: {0}'.format(e),HyperionWarning)


    def set_level(self, default_level):
        """
        Function to set the level for the logging object.
        :param default_level:
        :return: None
        """
        try:
            self.log.setLevel(logging_levels[default_level.lower()])

        except Exception as e:
            warnings.warn("Failed to set level: {0}".format(e),HyperionWarning)

    def set_format(self, format):
        try:
            logging_format = logging.Formatter(format)
            self.handler.setFormatter(logging_format)
        except Exception as e:
            warnings.warn("Failed to set format for logger: {0}".format(e),HyperionWarning)

    def debug(self, message):
        try:
            self.log.debug(message)
        except Exception as e:
            warnings.warn("Failed to deliver message: {}".format(e), HyperionWarning)

    def info(self, message):
        try:
            self.log.info(message)
        except Exception as e:
            warnings.warn("Failed to deliver message: {}".format(e), HyperionWarning)

    def warning(self, message):
        try:
            self.log.warning(message)
        except Exception as e:
            warnings.warn("Failed to deliver message: {}".format(e), HyperionWarning)

    def error(self, message):
        try:
            self.log.error(message)
        except Exception as e:
            warnings.warn("Failed to deliver message: {}".format(e), HyperionWarning)

    def critical(self, message):
        try:
            self.log.critical(message)
        except Exception as e:
            warnings.warn("Failed to deliver message: {}".format(e), HyperionWarning)


def get_logger():
    """
    function to grab a logger object with the cloudwatch handler attached.
    :param stream_name:
    :param log_group:
    :param default_level:
    :return: logger object
    """

    try:
        log = HyperionLogger()
        return log
    except Exception as e:
        warnings.warn("Unable to create logger because of: \n {}".format(e), HyperionWarning)


def _idempotent_create(_callable, *args, **kwargs):
    try:
        _callable(*args, **kwargs)
    except ClientError as e:
        if e.response.get("Error", {}).get("Code") != "ResourceAlreadyExistsException":
            raise


class HyperionWarning(UserWarning):
    pass


class CloudWatchLogHandler(handler_base_class):
    """
    Create a new CloudWatch log handler object. This is the main entry point to the functionality of the module. See
    http://docs.aws.amazon.com/AmazonCloudWatch/latest/DeveloperGuide/WhatIsCloudWatchLogs.html for more information.

    :param log_group: Name of the CloudWatch log group to write logs to. By default, the name of this module is used.
    :type log_group: String
    :param stream_name:
        Name of the CloudWatch log stream to write logs to. By default, the name of the logger that processed the
        message is used. Accepts a format string parameter of {logger_name}.
    :type stream_name: String
    :param use_queues:
        If **True**, logs will be queued on a per-stream basis and sent in batches. To manage the queues, a queue
        handler thread will be spawned.
    :type queue: Boolean
    :param send_interval:
        Maximum time (in seconds, or a timedelta) to hold messages in queue before sending a batch.
    :type send_interval: Integer
    :param max_batch_size:
        Maximum size (in bytes) of the queue before sending a batch. From CloudWatch Logs documentation: **The maximum
        batch size is 1,048,576 bytes, and this size is calculated as the sum of all event messages in UTF-8, plus 26
        bytes for each log event.**
    :type max_batch_size: Integer
    :param max_batch_count:
        Maximum number of messages in the queue before sending a batch. From CloudWatch Logs documentation: **The
        maximum number of log events in a batch is 10,000.**
    :type max_batch_count: Integer
    :param boto3_session:
        Session object to create boto3 `logs` clients. Accepts AWS credential, profile_name, and region_name from its
        constructor.
    :type boto3_session: boto3.session.Session
    :param create_log_group:
        Create log group.  **True** by default.
    :type create_log_group: Boolean
    """
    END = 1
    FLUSH = 2

    # extra size of meta information with each messages
    EXTRA_MSG_PAYLOAD_SIZE = 26

    @staticmethod
    def _get_session(boto3_session, boto3_profile_name):
        if boto3_session:
            return boto3_session

        if boto3_profile_name:
            return boto3.session.Session(profile_name=boto3_profile_name)

        return boto3

    def __init__(self, log_group=__name__, stream_name=None, use_queues=True, send_interval=60,
                 max_batch_size=1024 * 1024, max_batch_count=10000, boto3_session=None,
                 boto3_profile_name=None, create_log_group=True, *args, **kwargs):
        handler_base_class.__init__(self, *args, **kwargs)
        self.log_group = log_group
        self.stream_name = stream_name
        self.use_queues = use_queues
        self.send_interval = send_interval
        self.max_batch_size = max_batch_size
        self.max_batch_count = max_batch_count
        self.queues, self.sequence_tokens = {}, {}
        self.threads = []
        self.shutting_down = False

    def create_session(self, profile_name=None):
        try:
            self.cwl_client = boto3.session.Session(profile_name=profile_name, region_name='us-east-1').client('logs')
        except ClientError:
            print('Tokens have expired.')
        except NoRegionError:
            profile_name = '093937234853-GR_GG_COF_AWS_CoreEngineering'
            self.cwl_client = boto3.session.Session(profile_name=profile_name, region_name='us-east-1').client('logs')
        except ProfileNotFound:
            self.sts_client = boto3.client('sts')
            self.role = self.sts_client.assume_role(RoleARN='arn:aws:iam::093937234853:role/dms-cloudwatch-logs-role',
                                                    RoleSession='LogSession')
            self.credentials = self.role['Credentials']
            self.sts_client = boto3.client('logs',
                                           aws_access_key_id=self.credentials['AccessKeyId'],
                                           aws_secret_access_key=self.credentials['SecretAccessKey'],
                                           aws_session_token=self.credentials['SessionToken'],
                                           region_name='us-east-1'
                                           )
        except Exception as e:
            warnings.warn('Unable to create session: {0}'.format(e),HyperionWarning)

    def create_log_group(self):
        _idempotent_create(self.cwl_client.create_log_group,
                           logGroupName=self.log_group)

    def _submit_batch(self, batch, stream_name, max_retries=5):
        if len(batch) < 1:
            return
        sorted_batch = sorted(batch, key=itemgetter('timestamp'), reverse=False)
        kwargs = dict(logGroupName=self.log_group, logStreamName=stream_name,
                      logEvents=sorted_batch)
        if self.sequence_tokens[stream_name] is not None:
            kwargs["sequenceToken"] = self.sequence_tokens[stream_name]
        response = None

        for retry in range(max_retries):
            try:
                response = self.cwl_client.put_log_events(**kwargs)
                break
            except ClientError as e:
                if e.response.get("Error", {}).get("Code") in ("DataAlreadyAcceptedException",
                                                               "InvalidSequenceTokenException"):
                    kwargs["sequenceToken"] = e.response["Error"]["Message"].rsplit(" ", 1)[-1]
                else:
                    warnings.warn("Failed to deliver logs: {}".format(e), HyperionWarning)
            except Exception as e:
                warnings.warn("Failed to deliver logs: {}".format(e), HyperionWarning)

        # response can be None only when all retries have been exhausted
        if response is None or "rejectedLogEventsInfo" in response:
            warnings.warn("Failed to deliver logs: {}".format(response), HyperionWarning)

    def emit(self, message):
        stream_name = self.stream_name
        if stream_name is None:
            stream_name = message.name
        else:
            stream_name = stream_name.format(logger_name=message.name)
        if stream_name not in self.sequence_tokens:
            _idempotent_create(self.cwl_client.create_log_stream,
                               logGroupName=self.log_group, logStreamName=stream_name)
            self.sequence_tokens[stream_name] = None

        if isinstance(message.msg, collections.Mapping):
            message.msg = json.dumps(message.msg)

        cwl_message = dict(timestamp=int(message.created * 1000), message=self.format(message))

        if self.use_queues:
            if stream_name not in self.queues:
                self.queues[stream_name] = queue.Queue()
                thread = threading.Thread(target=self.batch_sender,
                                          args=(self.queues[stream_name], stream_name, self.send_interval,
                                                self.max_batch_size, self.max_batch_count))
                self.threads.append(thread)
                thread.daemon = True
                thread.start()
            if self.shutting_down:
                warnings.warn("Received message after logging system shutdown", HyperionWarning)
            else:
                self.queues[stream_name].put(cwl_message)
        else:
            self._submit_batch([cwl_message], stream_name)

    def batch_sender(self, my_queue, stream_name, send_interval, max_batch_size, max_batch_count):
        # thread_local = threading.local()
        msg = None

        def size(_msg):
            return (len(_msg["message"]) if isinstance(_msg, dict) else 1) + CloudWatchLogHandler.EXTRA_MSG_PAYLOAD_SIZE

        def truncate(_msg2):
            warnings.warn("Log message size exceeds CWL max payload size, truncated", HyperionWarning)
            _msg2["message"] = _msg2["message"][:max_batch_size - CloudWatchLogHandler.EXTRA_MSG_PAYLOAD_SIZE]
            return _msg2

        # See https://boto3.readthedocs.io/en/latest/reference/services/logs.html#CloudWatchLogs.Client.put_log_events
        while msg != self.END:
            cur_batch = [] if msg is None or msg == self.FLUSH else [msg]
            cur_batch_size = sum(size(msg) for msg in cur_batch)
            cur_batch_msg_count = len(cur_batch)
            cur_batch_deadline = time.time() + send_interval
            while True:
                try:
                    msg = my_queue.get(block=True, timeout=max(0, cur_batch_deadline - time.time()))
                    if size(msg) > max_batch_size:
                        msg = truncate(msg)
                except queue.Empty:
                    # If the queue is empty, we don't want to reprocess the previous message
                    msg = None
                if msg is None \
                        or msg == self.END \
                        or msg == self.FLUSH \
                        or cur_batch_size + size(msg) > max_batch_size \
                        or cur_batch_msg_count >= max_batch_count \
                        or time.time() >= cur_batch_deadline:
                    self._submit_batch(cur_batch, stream_name)
                    if msg is not None:
                        # We don't want to call task_done if the queue was empty and we didn't receive anything new
                        my_queue.task_done()
                    break
                elif msg:
                    cur_batch_size += size(msg)
                    cur_batch_msg_count += 1
                    cur_batch.append(msg)
                    my_queue.task_done()

    def flush(self):
        for q in self.queues.values():
            q.put(self.FLUSH)
        for q in self.queues.values():
            q.join()

    def close(self):
        self.shutting_down = True
        for q in self.queues.values():
            q.put(self.END)
        for q in self.queues.values():
            q.join()
        handler_base_class.close(self)
